
Key takeaway:
- Background is more powerful than popup in its awareness and accessibility. When setting badge text or badge background color in background.js, you do not need to pass the tab ID of the active tab. Background.js is already aware of the active tab so that parameter tabId becomes optional. So once we moved the code from popup.js into background.js, we could remove the tabId parameter and remove the storing of the tabId (via Storage API to have it persist into onAlarm).

---

## tabId not needed for background actions

We can simplify background.js because some of what's needed at popup.js isn't needed because background.js has increased awareness / abilities

We can simplify background.js to:
```
console.log("background.js loaded");

// At the start of your background.js
chrome.runtime.onStartup.addListener(async () => {
  // Restore badge state from storage
  const state = await chrome.storage.local.get(['badgeState']);
  if (state.badgeState) {
    if(state.badgeState.text) {
      chrome.action.setBadgeText({
        text: state.badgeState.text
      });
    } else if(state.badgeState.color) {
      chrome.action.setBadgeBackgroundColor({
        color: state.badgeState.color
      });
    }
  }
});

chrome.runtime.onMessage.addListener(async function (request, sender, sendResponse) {
  switch (request.type) {
    case "activeTab":
      const { tabId, href } = request;

      chrome.tabs.update(tabId, { url: href }, async (tabInfo) => {
        console.log('Tab updated, waiting for load...');
        chrome.action.setBadgeText({
          text: "WAIT",
        });
        chrome.action.setBadgeBackgroundColor({
          color: "#FF0000",
        });

        // Wait for the tab to complete loading
        await new Promise((resolve) => {
          chrome.tabs.onUpdated.addListener(function listener(updatedTabId, info) {
            if (updatedTabId === tabId && info.status === 'complete') {
              chrome.tabs.onUpdated.removeListener(listener);
              resolve();
            }
          });
        });

        try {
          console.log('Attempting to change favicon...');
          await chrome.scripting.executeScript({
            target: { tabId: tabId },
            func: () => {
              console.log('Script executing in tab...');

              // Change favicon
              let link = document.querySelector("link[rel*='icon']");
              if (!link) {
                link = document.createElement('link');
                link.rel = 'shortcut icon';
              }
              link.type = 'image/x-icon';
              link.href = 'https://cdn-icons-png.flaticon.com/512/16000/16000326.png';
              document.head.appendChild(link);

              // Change title
              document.title = "ACTIVE";

              // Add visual feedback
              const notification = document.createElement('div');
              notification.id = "favicon-changed-notification";
              notification.textContent = 'Favicon changed!';
              notification.style.cssText = 'position: fixed; top: 20px; right: 20px; background: #4CAF50; color: white; padding: 10px; border-radius: 5px; z-index: 9999;';
              document.body.appendChild(notification);
            }
          });

          console.log('notification-duration-reached before alarm created')
          // Set up alarm to clear notification and storage after 3 seconds, including the tabId
          chrome.alarms.create('notification-duration-reached', {
            delayInMinutes: 0.05, // 3 seconds
            periodInMinutes: 0
          });
          // Store the tabId for the alarm handler
          chrome.storage.local.set({ 'timerTabId': tabId });

          // When setting badge
          const text = "OK";
          const color = "#FFFFFF";
          await chrome.action.setBadgeText({text});
          await chrome.action.setBadgeBackgroundColor({color});
          const badgeState = { text, color };
          await chrome.storage.local.set({ badgeState }); // Store the state

        } catch (error) {
          console.error('Failed to execute script:', error);
          chrome.action.setBadgeText({
            text: "ERR",
          });
        }
      });
      break;
  } // switch
});

chrome.alarms.onAlarm.addListener((alarm) => {
  if (alarm.name === 'notification-duration-reached') {
    chrome.storage.local.get(['timerTabId'], function(result) {
      const tabId = result.timerTabId;
      if (!tabId) return;

	  // Get the stored tabId instead of querying active tab
      console.log('notification-duration-reached onAlarm')
      
      // Clear the badge
      chrome.action.setBadgeText({
        text: "",
      });

      // Remove the notification
      chrome.scripting.executeScript({
        target: { tabId: tabId },
        func: () => {
          try {
            const notification = document.getElementById("favicon-changed-notification");
            if (notification) {
              notification.remove();
            }
          } catch (error) {
            console.error('Error removing notification:', error);
          }
        }
      }).catch(error => {
        console.error('Failed to execute script:', error);
      });

      // Clean up stored tabId
      chrome.storage.local.remove('timerTabId');
    })
    }
});
```

What we've done is removed all `tabId` from `chrome.action.setBadgeText` and `chrome.action.setBadgeBackgroundColor`, at: 
- at the start of background.js where it checks for any existing badge to continue re-rendering (because if background service worker restarted, then previous badge visual aborted).
- at when the badge is set to a white "OK" after a red "WAIT" and the 3 second countdown is created
- at when the 3 second countdown is reached, then the badge is set to empty text ""

We also removed `tabId` from the badgeState that's saved to Storage to keep the badge information persistent into the alarm execution context.

This is all due to background.js Chrome action having more awareness. If you were running setBadgeText or setBadgeBackgroundColor anywhere else (like popup.js), you'd needed to provide the tabId.

We CANNOT remove `tabId` from `chrome.scripting.executeScript` because manifest v3 hasn't given that awareness to background.js.

### 🟨 Badge API Usage Comparison Across Script Contexts

|**Context**|`chrome.action.setBadgeText`|`chrome.action.setBadgeBackgroundColor`|Notes|
|---|---|---|---|
|**Background Scripts**|✅ **Recommended**; reliable & persistent|✅ **Recommended**; consistent updates|Can respond to events and persist state across sessions|
|**Popup Scripts**|⚠️ Works only when popup is open|⚠️ Works only when popup is open|Temporary; updates may disappear after popup closes|
|**Sidebar Pages**|⚠️ Same limits as popup|⚠️ Same limits as popup|Sidebar behaves like a persistent popup but still within extension UI|
|**Devtools Panel**|❌ Not allowed|❌ Not allowed|No access to `chrome.action.*` APIs in DevTools context|
|**Devtools Sidebar**|❌ Not allowed|❌ Not allowed|Same restriction as DevTools panel|
|**Content Scripts**|❌ Not allowed (no access to `chrome.action`)|❌ Not allowed|Must message background script to update badge|

---

### ✅ Recommendations

- **Always use background scripts** for badge updates when possible.
- **Use messaging** from content, devtools, or popup scripts to inform the background script to update the badge.

We've been doing just that when originally the badge logic was in popup.js but we've since moved that logic to background.js, and have popup.js message background.js when it's time to start visualizing with the badge (Chrome extension icon).

## Notifications system

We can add this below the green notification code at background.js to send a message a message typed "notifications":
```
chrome.runtime.sendMessage({ type: "notification" });
```

^ Why not jus replace the green notification? Because there are a lot of settings involved especially on Mac, so the user may not see your notification.

Then in the same background.js, we listen for the message, and upon hearing it, then we create the notification:.
```
chrome.runtime.onMessage.addListener(async function (request, sender, sendResponse) {
  switch (request.type) {
    case "notification":
      chrome.notifications.create(
        { type: "basic", iconUrl: "http://www.google.com/favicon.ico", title: "active", message: "Favicon changed successfully", priority: 0 },
        (notificationId) => {
          if (chrome.runtime.lastError) {
            console.error(chrome.runtime.lastError.message);
          }
        });
    
      chrome.notifications.getPermissionLevel((notificationPermissionLevel) => {
        console.log("Noifications level: " + notificationPermissionLevel);
      })
      break;
	case "activeTab":
	  // ...
```

At manifest.json, add the `"notifications"` permission to the `permissions` array:
- Not exhaustive:
```
"permissions": ["notifications"],
```

Now make sure notifications are allow between Chrome and your device.

1. Make sure your Chrome settings allow notifications:
chrome://settings/content

2. On Mac, Allow Notifications for Google Chrome, Google Chrome Dev, AND Google Chrome Alert and allow it to show in Notification Center:
![[Pasted image 20250322145824.png]]

^ Banner is a brief notification at the top right.
^ Alert is a notification at the top right that the user must dismiss.

You can open Notifications Center on the right to see previous notifications (Click date/time to open notifications drawer)
![[Pasted image 20250322145939.png]]

3. Make sure your device is not in DND mode. On Mac, even Allowed mode blocks notifications - Try turning off Focus or try turning on Personal.

4. On Mac, if currently mirroring or sharing the display, make sure you ticked to allow notifications when mirroring or sharing the display.

![[Pasted image 20250322153111.png]]

You should see a notification at top right on Mac:
![[Pasted image 20250322161526.png]]